package com.telepathicgrunt.the_bumblezone.items;

import com.telepathicgrunt.the_bumblezone.modinit.BzCriterias;
import com.telepathicgrunt.the_bumblezone.modinit.BzItems;
import com.telepathicgrunt.the_bumblezone.modinit.BzSounds;
import com.telepathicgrunt.the_bumblezone.modinit.BzTags;
import com.telepathicgrunt.the_bumblezone.platform.ItemExtension;
import com.telepathicgrunt.the_bumblezone.utils.OptionalBoolean;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.sounds.SoundSource;
import net.minecraft.stats.Stats;
import net.minecraft.world.InteractionHand;
import net.minecraft.world.InteractionResultHolder;
import net.minecraft.world.entity.EquipmentSlot;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.entity.projectile.AbstractArrow;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.ProjectileWeaponItem;
import net.minecraft.world.item.UseAnim;
import net.minecraft.world.item.Vanishable;
import net.minecraft.world.item.enchantment.Enchantment;
import net.minecraft.world.item.enchantment.EnchantmentHelper;
import net.minecraft.world.item.enchantment.Enchantments;
import net.minecraft.world.level.Level;
import net.minecraft.world.phys.Vec3;
import org.joml.Quaternionf;
import org.joml.Quaternionfc;
import org.joml.Vector3f;

import java.util.function.Predicate;

public class CrystalCannon extends ProjectileWeaponItem implements Vanishable, ItemExtension {
    public static final String TAG_CRYSTALS = "CrystalStored";

    public CrystalCannon(Properties properties) {
        super(properties.durability(80));
    }

    @Override
    public void releaseUsing(ItemStack crystalCannon, Level level, LivingEntity livingEntity, int currentDuration) {
        if (!level.isClientSide() && livingEntity instanceof Player player) {
            ItemStack mutableCrystalCannon = player.getItemInHand(InteractionHand.MAIN_HAND);

            int numberOfCrystals = getNumberOfCrystals(mutableCrystalCannon);
            int quickCharge = EnchantmentHelper.getItemEnchantmentLevel(Enchantments.QUICK_CHARGE, mutableCrystalCannon);
            int remainingDuration = this.getUseDuration(mutableCrystalCannon) - currentDuration;
            if (remainingDuration >= 20 - (quickCharge * 3) && numberOfCrystals > 0) {
                int crystalsToSpawn = getAndClearStoredCrystals(level, mutableCrystalCannon);
                for (int i = 0; i < crystalsToSpawn; i++) {
                    float offset = 0;
                    if (i == 1) {
                        offset = (livingEntity.getRandom().nextFloat() * 5f) + 3.5f;
                    }
                    else if (i == 2) {
                        offset = (livingEntity.getRandom().nextFloat() * 5f) - 11.5f;
                    }
                    else if (i != 0) {
                        offset = livingEntity.getRandom().nextFloat() * 10f - 5f;
                    }

                    AbstractArrow newCrystal = BzItems.HONEY_CRYSTAL_SHARDS.get().createArrow(level, crystalCannon, livingEntity);

                    double weaponDamage = newCrystal.getBaseDamage();
                    int power = EnchantmentHelper.getItemEnchantmentLevel(Enchantments.POWER_ARROWS, crystalCannon);
                    if (power > 0) {
                        weaponDamage += (power * 0.5D) + 0.5D;
                    }
                    newCrystal.setBaseDamage(weaponDamage);

                    int punch = EnchantmentHelper.getItemEnchantmentLevel(Enchantments.PUNCH_ARROWS, crystalCannon);
                    if (punch > 0) {
                        newCrystal.setKnockback(newCrystal.getKnockback() + punch);
                    }

                    int pierce = EnchantmentHelper.getItemEnchantmentLevel(Enchantments.PIERCING, crystalCannon);
                    if (pierce > 0) {
                        newCrystal.setPierceLevel((byte) pierce);
                    }

                    newCrystal.pickup = AbstractArrow.Pickup.CREATIVE_ONLY;
                    Vec3 playerEyePos = new Vec3(
                            player.getX(),
                            player.getEyeY() - 0.25f,
                            player.getZ());
                    newCrystal.moveTo(playerEyePos.x(),
                            playerEyePos.y(),
                            playerEyePos.z(),
                            player.getYRot(),
                            player.getXRot());

                    Vec3 upVector = player.getUpVector(1.0F);
                    Vec3 viewVector = player.getViewVector(1.0F);
                    Vector3f shootVector = viewVector.toVector3f();
                    if (i != 0) {
                        Quaternionfc quaternion1 = new Quaternionf(upVector.x(), upVector.y(), upVector.z(), offset);
                        shootVector.rotate(quaternion1);
                    }

                    float weaponProjectileSpeed = 1.9F;
                    newCrystal.shoot(
                            shootVector.x(),
                            shootVector.y() + (livingEntity.getRandom().nextFloat() * 0.2f + 0.01f),
                            shootVector.z(),
                            weaponProjectileSpeed,
                            1);
                    level.addFreshEntity(newCrystal);

                    player.awardStat(Stats.ITEM_USED.get(mutableCrystalCannon.getItem()));

                    level.playSound(null, player.blockPosition(), BzSounds.CRYSTAL_CANNON_FIRES.get(), SoundSource.PLAYERS, 1.0F, (player.getRandom().nextFloat() * 0.2F) + 0.6F);
                    mutableCrystalCannon.hurtAndBreak(1, player, playerEntity -> playerEntity.broadcastBreakEvent(EquipmentSlot.MAINHAND));

                    if(numberOfCrystals >= 3 && player instanceof ServerPlayer serverPlayer) {
                        BzCriterias.CRYSTAL_CANNON_FULL_TRIGGER.get().trigger(serverPlayer);
                    }
                }

                // Consume one extra durability
                mutableCrystalCannon.hurtAndBreak(1, player, playerEntity -> playerEntity.broadcastBreakEvent(EquipmentSlot.MAINHAND));
            }
        }
    }

    @Override
    public InteractionResultHolder<ItemStack> use(Level level, Player player, InteractionHand interactionHand) {
        ItemStack crystalCannon = player.getItemInHand(interactionHand);
        if (!level.isClientSide()) {
            loadProjectiles(player, crystalCannon);
        }

        if (getNumberOfCrystals(crystalCannon) == 0) {
            return InteractionResultHolder.fail(crystalCannon);
        }
        else {
            player.startUsingItem(interactionHand);
            return InteractionResultHolder.consume(crystalCannon);
        }
    }

    private void loadProjectiles(Player player, ItemStack crystalCannon) {
        ItemStack projectItem1 = player.getProjectile(crystalCannon);
        if (projectItem1.isEmpty()) {
            return;
        }

        if (tryAddCrystal(crystalCannon)) {
            boolean infinite = player.getAbilities().instabuild ||
                    (projectItem1.getItem() instanceof HoneyCrystalShards &&
                            ((HoneyCrystalShards)projectItem1.getItem()).bz$isInfinite(projectItem1, crystalCannon, player).orElseGet(() -> false));

            if (!infinite) {
                projectItem1.shrink(1);
                if (projectItem1.isEmpty()) {
                    player.getInventory().removeItem(projectItem1);
                }
            }
        }
    }

    public static int getAndClearStoredCrystals(Level level, ItemStack crystalCannonItem) {
        int numberOfCrystals = getNumberOfCrystals(crystalCannonItem);
        if(numberOfCrystals > 0) {
            CompoundTag tag = crystalCannonItem.getOrCreateTag();
            int crystals = tag.getInt(TAG_CRYSTALS);
            tag.putInt(TAG_CRYSTALS, Math.max(0, crystals - 3));
        }
        return numberOfCrystals;
    }

    public static boolean tryAddCrystal(ItemStack crystalCannonItem) {
        if(getNumberOfCrystals(crystalCannonItem) < 3) {
            CompoundTag tag = crystalCannonItem.getOrCreateTag();
            int crystals = tag.getInt(TAG_CRYSTALS);
            tag.putInt(TAG_CRYSTALS, crystals + 1);
            return true;
        }
        return false;
    }

    public static int getNumberOfCrystals(ItemStack crystalCannonItem) {
        if(crystalCannonItem.hasTag()) {
            CompoundTag tag = crystalCannonItem.getOrCreateTag();
            return tag.getInt(TAG_CRYSTALS);
        }
        return 0;
    }

    /**
     * Return whether this item is repairable in an anvil.
     */
    @Override
    public boolean isValidRepairItem(ItemStack toRepair, ItemStack repair) {
        return repair.is(BzTags.CRYSTAL_CANNON_REPAIR_ITEMS);
    }

    @Override
    public Predicate<ItemStack> getSupportedHeldProjectiles() {
        return (itemStack) -> itemStack.is(BzItems.HONEY_CRYSTAL_SHARDS.get());
    }

    @Override
    public Predicate<ItemStack> getAllSupportedProjectiles() {
        return (itemStack) -> itemStack.is(BzItems.HONEY_CRYSTAL_SHARDS.get());
    }

    @Override
    public int getEnchantmentValue() {
        return 1;
    }

    @Override
    public int getDefaultProjectileRange() {
        return 0;
    }

    @Override
    public int getUseDuration(ItemStack itemStack) {
        return 72000;
    }

    @Override
    public UseAnim getUseAnimation(ItemStack itemStack) {
        return UseAnim.BOW;
    }

    @Override
    public OptionalBoolean bz$canApplyAtEnchantingTable(ItemStack stack, Enchantment enchantment) {
        if(enchantment == Enchantments.QUICK_CHARGE ||
            enchantment == Enchantments.PIERCING ||
            enchantment == Enchantments.POWER_ARROWS ||
            enchantment == Enchantments.PUNCH_ARROWS)
        {
            return OptionalBoolean.TRUE;
        }

        return OptionalBoolean.of(enchantment.category.canEnchant(stack.getItem()));
    }

    // Runs on Forge
    public boolean canApplyAtEnchantingTable(ItemStack stack, Enchantment enchantment) {
        return this.bz$canApplyAtEnchantingTable(stack, enchantment)
                .orElseGet(() -> enchantment.category.canEnchant(stack.getItem()));
    }
}
